# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow** contains only ``__version__`` to be updated, nothing more.
This is in an attempt to make this module easy to load initially.
"""

__version__ = "0.24.0"
