// Copyright (c) {{COPY.YEAR}} {{COPY.HOLDER}}
// This code is licensed under {{COPY.LICENSE}} license (see LICENSE for details)

{{#with.conan}}
#include <fmt/format.h>
#include <gtest/gtest.h>
{{/with.conan}}

namespace {{NAMESPACE}}::testing {

{{#with.conan}}
	TEST({{NAMESPACE}}, test) { ASSERT_EQ(1, 1); }
{{/with.conan}}

}  // namespace {{NAMESPACE}}::testing
